import argparse
import sys
from pathlib import Path

import nbformat
import nbparameterise
import yaml

from .notebooks import notebooks

PKG_DIR = Path(__file__).parent
ACTIONS = ("CORRECT", "DARK")  # Only these use this config system


def params_from_nb(rel_path):
    abs_path = PKG_DIR / rel_path
    nb = nbformat.read(abs_path, as_version=4)
    return nbparameterise.code.extract_parameter_dict(nb)


def load_nb_params():
    res = {}
    n = 0
    for detector, det_nbs in notebooks.items():
        res[detector] = det_params = {}
        for action in ACTIONS:
            if action not in det_nbs:
                continue

            det_params[action] = params_from_nb(det_nbs[action]["notebook"])
            n += 1

            for dep_nb in det_nbs[action].get("dep_notebooks", []):
                dep_params = params_from_nb(dep_nb)
                extras = set(dep_params) - set(det_params[action])
                if extras:
                    print(f"Extra parameters in dep notebook for {detector} {action}")
                    print(" ", dep_nb)
                    print(" ", sorted(extras))

    print(f"Loaded parameters for {n} correct/dark notebooks\n")
    return res

n_probs = 0

def config_problem(problem, detail, file, key):
    global n_probs
    n_probs += 1
    print(f"{problem}\n  {detail}\n  file: {file}\n  key: {'/'.join(key)}\n")


def bad_list_element_types(value, param: nbparameterise.Parameter):
    # param.type == list
    nb_types = {type(e) for e in param.value}
    if nb_types == {float, int}:
        nb_types = {float}
    assert len(nb_types) == 1, f"{param.value}"
    nb_type = nb_types.pop()

    if not isinstance(value, list):
        value = [value]  # In the xfel-calibrate CLI, a list of 1 item is not distinct

    if "RANGE ALLOWED" in (param.comment or "").upper():
        new_value = []
        # Try to parse ranges
        for e in value:
            if not isinstance(e, str):
                new_value.append(e)
                continue
            try:
                for rcomp in e.split(","):
                    if "-" in rcomp:
                        start, end = rcomp.split("-")
                        new_value += list(range(int(start), int(end)))
                    else:
                        new_value += [int(rcomp)]
            except ValueError:
                return "Invalid range specified", repr(e)
        value = new_value

    if isinstance(value, list):
        if any(not like_type(e, nb_type) for e in value):
            bad_types = sorted({type(e) for e in value})
            return ("List element type differs from default in notebook",
                    f"notebook: {nb_type}, config: {bad_types}")
    return None, None


def like_type(value, param_type):
    if isinstance(value, param_type):
        return True
    if param_type is float and isinstance(value, int):
        return True
    return False


def check_yaml_file(path: Path, config_dir: Path, nb_params, detectors):
    rel_path = path.relative_to(config_dir)
    with path.open('r') as f:
        cfg = yaml.safe_load(f)

    kid_to_det = {}
    for kid, d in cfg['data-mapping'].items():
        det = d['detector-type'].upper()
        if det in nb_params:
            kid_to_det[kid] = det
        else:
            config_problem(
                "Unknown detector type in data mapping", det,
                rel_path, ('data-mapping', kid, 'detector-type')
            )

    for action in ACTIONS:
        action = action.lower()
        for instrument in cfg[action]:
            inst_cfg = cfg[action][instrument]
            for kid, det_cfg in inst_cfg.items():
                try:
                    det = kid_to_det[kid]
                except KeyError:
                    config_problem(
                        "Detector Karabo ID not in data mapping", kid,
                        rel_path, (action, instrument)
                    )
                    continue

                if detectors and (det not in detectors):
                    continue

                try:
                    det_params = nb_params[det][action.upper()]
                except KeyError:
                    config_problem(
                        f"Action not available for detector {det}", action,
                        rel_path, (action, instrument, kid)
                    )
                    continue

                for k, v in (det_cfg or {}).items():
                    pname = k.replace('-', '_')
                    try:
                        param = det_params[pname]
                    except KeyError:
                        config_problem(
                            "Parameter name not found in notebook", pname,
                            rel_path, (action, instrument, kid)
                        )
                        continue

                    if param.type is list:
                        pblm, detail = bad_list_element_types(v, param)
                        if pblm:
                            config_problem(
                                pblm, detail, rel_path, (action, instrument, kid, k)
                            )
                    elif not like_type(v, param.type):
                        config_problem(
                            "Type differs from default in notebook",
                            f"notebook: {param.value!r}, config: {v!r}",
                            rel_path, (action, instrument, kid, k)
                        )



def main(argv=None):
    ap = argparse.ArgumentParser()
    ap.add_argument('--config-dir', type=Path, default="~/calibration_config/")
    ap.add_argument('--detector', choices=notebooks.keys(), action='append')
    args = ap.parse_args(argv)

    config_dir = args.config_dir.expanduser()
    if not config_dir.is_dir():
        sys.exit(f"No such directory: {config_dir}")

    nb_params = load_nb_params()

    files = [config_dir / 'default.yaml'] + list(config_dir.glob('*/*.yaml'))
    print(f"Checking {len(files)} files in {config_dir}...\n")

    for file in files:
        check_yaml_file(file, config_dir, nb_params, detectors=args.detector)

    print(f"{n_probs} config issues found")
    sys.exit(1 if n_probs else 0)


if __name__ == "__main__":
    main()
